'use strict';

var async = require('async');
var request = require('request');
var dot = require('dot');
dot.templateSettings.strip = false;
var debug = require('../util/debug')('batch-client');

function BatchClient (endpoint, username, apiKey, hostHeaderTemplate, inlineExecution, queryRunner) {
    this.endpoint = endpoint;
    this.username = username;
    this.apiKey = apiKey;
    this.hostHeader = dot.template(hostHeaderTemplate || '{{=it.username}}.localhost.lan');

    this.inlineExecution = inlineExecution || false;
    this.queryRunner = queryRunner;
}

module.exports = BatchClient;

function batchErrorMessage (err, response) {
    if (err && err.message) {
        return err.message;
    }
    if (response.statusCode === 400 &&
        response.body && response.body.error &&
        response.body.error.some(e => e.match(/\bYour payload is too large\b/i))) {
        return 'The analysis SQL code couldn\'t be queued for execution because it is too complex. ' +
            'If you have multiple analyses please try simplifying ' +
            'or materialising some of them to reduce the payload.';
    }
    return 'Unable to enqueue SQL API batch job';
}

BatchClient.prototype.enqueue = function (queries, callback) {
    if (this.inlineExecution === true) {
        return runQueries(this.queryRunner, queries, callback);
    }

    var enqueueRequest = {
        json: true,
        url: this.endpoint,
        qs: { api_key: this.apiKey },
        headers: {
            Host: this.hostHeader({ username: this.username }),
            'User-Agent': 'camshaft'
        },
        body: {
            query: {
                query: queries
            }
        }
    };
    request.post(enqueueRequest, function (err, response, job) {
        response = response || {};
        if (err || response.statusCode !== 201) {
            return callback(new Error(batchErrorMessage(err, response)));
        }

        debug('Queued job: %j', job);

        return callback(null, {
            status: 'ok',
            job_id: job.job_id
        });
    });
};

BatchClient.prototype.getJob = function (jobId, callback) {
    var requestParams = {
        json: true,
        url: this.endpoint + '/' + jobId,
        qs: { api_key: this.apiKey },
        headers: {
            Host: this.hostHeader({ username: this.username }),
            'User-Agent': 'camshaft'
        }
    };

    request.get(requestParams, function (err, response, job) {
        if (err) {
            return callback(err);
        }

        if (response.statusCode === 404) {
            return callback();
        }

        if (response.statusCode === 400 && response.body.error.indexOf('not found') === -1) {
            return callback();
        }

        if (response.statusCode !== 200) {
            return callback(new Error('Unable to get SQL API batch job'));
        }

        debug('job: %j', job);

        callback(null, job);
    });
};

function runQueries (queryRunner, queries, callback) {
    var readOnlyQuery = true;
    var writeQuery = !readOnlyQuery;

    if (!queryRunner || !queryRunner.run) {
        return callback(new Error('BatchClient inline execution requires a valid QueryRunner instance'));
    }

    var jobId = '646f6e65-772f-3c33-6279-726f63686f61';

    async.eachSeries(
        queries,
        function (query, done) {
            if (typeof query === 'string') {
                return queryRunner.run(query, writeQuery, done);
            }

            queryRunner.run(query.query, writeQuery, function (err) {
                if (err) {
                    if (query.onerror) {
                        var onerrorQuery = query.onerror.replace(/<%=\s*job_id\s*%>/g, jobId);
                        onerrorQuery = onerrorQuery.replace(/<%=\s*error_message\s*%>/g, err.message);
                        return queryRunner.run(onerrorQuery, writeQuery, done);
                    }

                    return done(err);
                }

                if (query.onsuccess) {
                    var onsuccessQuery = query.onsuccess.replace(/<%=\s*job_id\s*%>/g, jobId);
                    return queryRunner.run(onsuccessQuery, writeQuery, done);
                }

                done();
            });
        },
        function finish (err, results) {
            if (err) {
                debug('Unable to run query. Reason: %s', err.message);
                if (queries.onerror) {
                    return queryRunner.run(queries.onerror, writeQuery, function () {
                        callback(new Error('Unable to run query'));
                    });
                }
                return callback(new Error('Unable to run query'));
            }

            if (queries.onsuccess) {
                return queryRunner.run(queries.onsuccess, writeQuery, function () {
                    callback(null, results);
                });
            }

            return callback(null, results);
        }
    );
}
